#define WANTVXDWRAPS

#include <dos.h>
#include <basedef.h>
#include <vmm.h>
#include <vpicd.h>
#include <vxdwraps.h>
#include <vwin32.h>
#include <winerror.h>
#include "a1216e.h"

#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG

DWORD cbIn, cbOut; // Sizes of input and output buffers
unsigned long *lpOut; // Pointers to input and output buffers

unsigned char IRQ; // holds irq sent by dll
unsigned long BaseAddress; // holds base address sent by dll
short *ChannelTable; // points to the array of channels to acquire, terminated with -1
unsigned char GainTable[256]; // points to the array of gains (128 total)
unsigned long *DataBuffer; // points to the buffer to hold the counts (low 2 bytes) and ch # (hi 2 bytes)
long NumPoints; // number of points requested by user
long PointsRemaining; // points left to acquire, on zero exits
long Index, NumChannelEntries;
BOOL IRQ_Hooked, IntDisabled, ChannelTableAllocated, DataBufferAllocated;
BOOL IRQ_Cleared;

HIRQ    hIRQ;
VID     A1216E_IRQ_Desc = {0,0,0,0,0,0,0,0,0};

unsigned short _stdcall inport(unsigned short port)
{
    unsigned short t=0;
    _asm {                            
        mov     dx,port
        in      ax,dx
        mov     t,ax
        }
    return t;
}
unsigned char _stdcall inportb(unsigned short port)
{
    unsigned short t=0;
    _asm {
        mov     dx,port
        in      al,dx
        xor     ah,ah
        mov     t,ax
        }
    return t;
}        
void _stdcall outport(unsigned short port, unsigned short val)
{
    _asm {
        mov     dx, port
        mov     ax, val
        out     dx, ax
        }
}
void _stdcall outportb(unsigned short port, unsigned char val)
{
    _asm {
        mov     dx, port
        mov     al, val
        out     dx, al
        }
}

/****************************************************************************
                  A1216E_DeviceIOControl
        // Returning a positive value will cause the WIN32 DeviceIOControl
        // call to return FALSE, the error code can then be retrieved
        // via the WIN32 API GetLastError.
****************************************************************************/
DWORD _stdcall A1216E_DeviceIOControl(DWORD  dwService,
                                        DWORD  dwDDB,
                                        DWORD  hDevice,
                                        LPDIOC lpDIOCParms)
{
    DWORD dwRetVal = 0;

    switch (dwService)
	{
		case DIOC_OPEN:		// same as GET_VERSION
			dwRetVal = 0;
			break;

		case DIOC_CLOSEHANDLE:
			dwRetVal = A1216E_CleanUp();
			break;

		case InitDriver:
			dwRetVal = A1216E_InitDriver(dwDDB, hDevice, lpDIOCParms);
			break;

		case InitChannelTable:
			dwRetVal = A1216E_InitChannelTable(dwDDB, hDevice, lpDIOCParms);
			break;

		case InitGainTable:
			dwRetVal = A1216E_InitGainTable(dwDDB, hDevice, lpDIOCParms);
			break;

		case Status:
			dwRetVal = A1216E_Status(dwDDB, hDevice, lpDIOCParms);
			break;

		case Stop:
			dwRetVal = A1216E_Stop(dwDDB, hDevice, lpDIOCParms);
			break;

		case GetData:
			dwRetVal = A1216E_GetData(dwDDB, hDevice, lpDIOCParms);
			break;

		default:
	        dwRetVal = ERROR_NOT_SUPPORTED;
			break;
	}
    return(dwRetVal);
}

DWORD _stdcall A1216E_InitGainTable(DWORD dwDDB, DWORD hDevice, LPDIOC lpDIOCtl)
{
	unsigned char *pBuffer;
	int i;

    pBuffer = (unsigned char *)lpDIOCtl->lpvInBuffer;
    cbIn = lpDIOCtl->cbInBuffer;

	for (i=0; i<sizeof(GainTable); i++)
		GainTable[i] = 0;

	for (i=0; i<(cbIn<sizeof(GainTable)?cbIn:sizeof(GainTable)); i++)
		GainTable[i] = pBuffer[i];

	*(PDWORD)(lpDIOCtl->lpcbBytesReturned) = 0;
    return(0);
}

DWORD _stdcall A1216E_InitChannelTable(DWORD dwDDB, DWORD hDevice, LPDIOC lpDIOCtl)
{
	short *pBuffer, *newTable;
	long n = 0, i = 0;

    pBuffer = (short *)lpDIOCtl->lpvInBuffer;
    cbIn = lpDIOCtl->cbInBuffer;
	
	// allocate memory to store copy of requested channels
	if (ChannelTable != 0)
		_HeapFree(ChannelTable, 0);
	ChannelTableAllocated = FALSE;
	ChannelTable = (short*)_HeapAllocate(cbIn, HEAPZEROINIT);
	if (ChannelTable == 0)
		return ERROR_NOT_ENOUGH_MEMORY;
	ChannelTableAllocated = TRUE;

	// copy customer's channels to our memory
	n = cbIn / 2;
	NumChannelEntries = 0;
	while (NumChannelEntries < n)
	{
		ChannelTable[NumChannelEntries] = pBuffer[NumChannelEntries];
		if (ChannelTable[NumChannelEntries] < 0)
			break;
		else
			NumChannelEntries++;
	}
	if (NumChannelEntries == 0)
	{
		_HeapFree(ChannelTable, 0);
		ChannelTableAllocated = FALSE;
		return ERROR_INVALID_PARAMETER;
	}

	// if there are entries after -1, resize our memory to optimize
	if (n - 1 > NumChannelEntries)
	{
		newTable = (short*)_HeapReAllocate(ChannelTable, (NumChannelEntries + 1) * 2, 0);
		if (newTable != 0)
			ChannelTable = newTable;
	}
	
	*(PDWORD)(lpDIOCtl->lpcbBytesReturned) = 0;
	return(0);
}

DWORD _stdcall A1216E_InitDriver(DWORD dwDDB, DWORD hDevice, LPDIOC lpDIOCtl)
{
	PA1216E_INIT_DATA pBuffer;

    pBuffer = (PA1216E_INIT_DATA)lpDIOCtl->lpvInBuffer;
    cbIn = lpDIOCtl->cbInBuffer;
	if (!ChannelTableAllocated)
		return ERROR_NOT_SUPPORTED;
	if (cbIn < sizeof(A1216E_INIT_DATA))
	{
		_HeapFree(ChannelTable, 0);
		ChannelTableAllocated = FALSE;
		return ERROR_INVALID_PARAMETER;
	}

    if (IRQ_Hooked)
        VPICD_Force_Default_Behavior(hIRQ);

	IRQ_Hooked = FALSE;

	// initialize card data
    IRQ = pBuffer->IRQ;
	BaseAddress = pBuffer->BaseAddress;
	NumPoints = pBuffer->NumPoints;
	PointsRemaining = NumPoints;

	if (NumPoints < 1)
	{
		_HeapFree(ChannelTable, 0);
		ChannelTableAllocated = FALSE;
		return ERROR_INVALID_PARAMETER;
	}

	// allocate buffer to store data
	if (DataBuffer != 0)
		_HeapFree(DataBuffer, 0);
	DataBufferAllocated = FALSE;
	DataBuffer = (unsigned long*)_HeapAllocate(NumPoints * 4, HEAPZEROINIT);
	if (DataBuffer == 0)
	{
		_HeapFree(ChannelTable, 0);
		ChannelTableAllocated = FALSE;
		return ERROR_NOT_ENOUGH_MEMORY;
	}
	DataBufferAllocated = TRUE;

	outportb(BaseAddress + 0xF, 0xB4); // stop counters
	outportb(BaseAddress, 0xEA); // set up command register
	IntDisabled = FALSE;
	inportb(BaseAddress + 1);

	// on powerup, the a1216e is asserting its IRQ line and the first IRQ
	// is usually invalid. the following will clear the pending IRQ.
    A1216E_IRQ_Desc.VID_IRQ_Number = IRQ;
    A1216E_IRQ_Desc.VID_Hw_Int_Proc = (ULONG)&A1216E_DummyISR;
	hIRQ = VPICD_Virtualize_IRQ(&A1216E_IRQ_Desc);
    VPICD_Physically_Unmask(hIRQ);
    VPICD_Force_Default_Behavior(hIRQ);

    A1216E_IRQ_Desc.VID_IRQ_Number = IRQ;
    A1216E_IRQ_Desc.VID_Hw_Int_Proc = (ULONG)&A1216E_ConversionReady;

    hIRQ = VPICD_Virtualize_IRQ(&A1216E_IRQ_Desc);

    IRQ_Hooked = TRUE;

    VPICD_Physically_Unmask(hIRQ);

	// set the first channel in the table
	Index = 0;
	A1216E_SetChannel(ChannelTable[Index]);

	*(PDWORD)(lpDIOCtl->lpcbBytesReturned) = 0;
    return(0);
}

DWORD _stdcall A1216E_GetData(DWORD dwDDB, DWORD hDevice, LPDIOC lpDIOCtl)
{
	long i;
	const int Boundary = 8;

//    if (!IRQ_Hooked)
//        return ERROR_NOT_SUPPORTED;

    lpOut = (unsigned long*)lpDIOCtl->lpvOutBuffer;
    cbOut = lpDIOCtl->cbOutBuffer;
	if (cbOut < NumPoints * Boundary)
		return ERROR_INVALID_PARAMETER;

	for (i=0; i<NumPoints; i++)
		lpOut[i*Boundary/sizeof(DataBuffer[0])] = DataBuffer[i];

	*(PDWORD)(lpDIOCtl->lpcbBytesReturned) = NumPoints * Boundary;
    return(0);
}

DWORD _stdcall A1216E_Status(DWORD dwDDB, DWORD hDevice, LPDIOC lpDIOCtl)
{
	long *pBuffer;

	pBuffer = (long*)lpDIOCtl->lpvOutBuffer;
    cbOut = lpDIOCtl->cbOutBuffer;
	if (cbOut < sizeof(long))
		return ERROR_INVALID_PARAMETER;

	*pBuffer = PointsRemaining;

    *(PDWORD)(lpDIOCtl->lpcbBytesReturned) = sizeof(long);
	return (0);
}

DWORD _stdcall A1216E_Stop(DWORD dwDDB, DWORD hDevice, LPDIOC lpDIOCtl)
{
	PointsRemaining = 0;
    *(PDWORD)(lpDIOCtl->lpcbBytesReturned) = 0;
	return (0);
}

void _stdcall A1216E_SetChannel(long Ch)
{
	unsigned char Value;

	Value  = (Ch & 0xF);  // set MA0-MA3 (a/d channel)
	Value |= (GainTable[Ch] & 0x3) << 4;  // set GN0-GN1 (gain)

	outportb(BaseAddress + 2, Value);
}

// ISR
void _cdecl A1216E_ConversionReady(void)
{
	unsigned long Value, Counts;

//    _asm sti;    // enable interrupts

	if (IntDisabled)
	{
		inport(BaseAddress + 1);
		VPICD_Phys_EOI(hIRQ);
		return;
	}

	Counts = ((unsigned long)inport(BaseAddress + 6)) >> 4;
	Value = Counts;
	Value |= (ChannelTable[Index] << 16);

	DataBuffer[NumPoints - PointsRemaining] = Value;

	if (--PointsRemaining < 1)
	{
		outportb(BaseAddress + 0xF, 0xB4); // stop counters
		outportb(BaseAddress, 0);
		IntDisabled = TRUE;
		inportb(BaseAddress + 1); // clear latch
		VPICD_Phys_EOI(hIRQ);
		return;
	}

	Index++;
	if (ChannelTable[Index] == -1)
		Index = 0;
	A1216E_SetChannel(ChannelTable[Index]);

	inportb(BaseAddress + 1); // clear latch
	VPICD_Phys_EOI(hIRQ);
}

// DUMMY ISR
void _cdecl A1216E_DummyISR(void)
{
//    _asm sti;    // enable interrupts
    VPICD_Phys_EOI(hIRQ);
	inportb(BaseAddress + 1); // clear latch
	IRQ_Cleared = TRUE;
}

DWORD _stdcall A1216E_CleanUp(void)
{
    if (IRQ_Hooked)
        VPICD_Force_Default_Behavior(hIRQ);

	_HeapFree(ChannelTable, 0);
	_HeapFree(DataBuffer, 0);
	ChannelTableAllocated = DataBufferAllocated = FALSE;

    return(VXD_SUCCESS);
}

DWORD _stdcall A1216E_Dynamic_Exit(void)
{
    return(VXD_SUCCESS);
}

DWORD _stdcall MyPageLock(DWORD lpMem, DWORD cbSize)
{
    DWORD LinPageNum, LinOffset, nPages;

    LinOffset = lpMem & 0xfff; // page offset of memory to map
    LinPageNum = lpMem >> 12;  // generate page number

    // Calculate # of pages to map globally
    nPages = ((lpMem + cbSize) >> 12) - LinPageNum + 1;

    //
    // Return global mapping of passed in pointer, as this new pointer
    // is how the memory must be accessed out of context.
    //
    return (_LinPageLock(LinPageNum, nPages, PAGEMAPGLOBAL) + LinOffset);
}

void _stdcall MyPageUnlock(DWORD lpMem, DWORD cbSize)
{
    DWORD LinPageNum, nPages;

    LinPageNum = lpMem >> 12;
    nPages = ((lpMem + cbSize) >> 12) - LinPageNum + 1;

    // Free globally mapped memory
    _LinPageUnlock(LinPageNum, nPages, PAGEMAPGLOBAL);
}

#pragma VxD_ICODE_SEG
#pragma VxD_IDATA_SEG

DWORD _stdcall A1216E_Dynamic_Init(void)
{
    return(VXD_SUCCESS);
}



